LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE IEEE.STD_LOGIC_MISC.ALL;

-- A single M512 can make a 16 word, 5-bit wide CAM.  This script can be parameterized to chain them together into wider/deeper CAMs.
-- Please keep the depth a multiple of 16, and the width a multiple of 5.  If you don't need a width that is a multiple of 5, just tie off the extra
-- bits so they always match(both when writing and matching)

ENTITY M512_CAM IS
GENERIC (
	NUM_WORDS : INTEGER := 128; -- Must be multiple of 16
	WRADDR_WIDTH : INTEGER := 7; -- Enter log2(NUM_WORDS).
	WIDTH : INTEGER := 15 -- Must be multiple of 5

);	
PORT
	(
	-- Write Ports
	wrdata			: IN	std_logic_vector(WIDTH-1 DOWNTO 0);
	wraddress		: IN	std_logic_vector(WRADDR_WIDTH-1 DOWNTO 0);
	wrdelete_n		: IN	std_logic;
	wren			: IN	std_logic;
	wrclock			: IN	std_logic;
	
	-- Match Ports
	matchdata		: IN	std_logic_vector(WIDTH-1 DOWNTO 0);
	matchen			: IN	std_logic;
	matchclock		: IN	std_logic;	
	matchaddr		: OUT 	std_logic_vector(WRADDR_WIDTH-1 DOWNTO 0); -- encoded output.  Remove this if unnecessary, as lot of logic needed to encode.
	match_onehot	: BUFFER	std_logic_vector(NUM_WORDS-1 DOWNTO 0); -- one-hot output
	match			: OUT	std_logic -- Determines if there is a match
	);
end M512_CAM ;

ARCHITECTURE arch OF M512_CAM IS

component M512
	PORT
	(
		data		: IN STD_LOGIC_VECTOR (0 DOWNTO 0);
		wren		: IN STD_LOGIC  := '1';
		wraddress		: IN STD_LOGIC_VECTOR (8 DOWNTO 0);
		rdaddress		: IN STD_LOGIC_VECTOR (4 DOWNTO 0);
		rden		: IN STD_LOGIC  := '1';
		wrclock		: IN STD_LOGIC ;
		rdclock		: IN STD_LOGIC ;
		q		: OUT STD_LOGIC_VECTOR (15 DOWNTO 0)
	);
end component;


CONSTANT NUM_M512S_DEPTH : INTEGER := NUM_WORDS/16;
CONSTANT NUM_M512S_WIDTH : INTEGER := WIDTH/5;

SIGNAL data: std_logic_vector(0 DOWNTO 0);
SIGNAL address: std_logic_vector(WIDTH+4 DOWNTO 0);
SIGNAL m512_wren : STD_LOGIC_VECTOR(NUM_M512S_DEPTH-1 DOWNTO 0);
TYPE single_match_array IS ARRAY (NUM_M512S_WIDTH-1 DOWNTO 0) OF STD_LOGIC_VECTOR(NUM_WORDS-1 DOWNTO 0); 
SIGNAL single_match_onehot : single_match_array;

begin

	data(0) <= wrdelete_n;

instantiate_M512s_depth : FOR i IN 0 TO NUM_M512S_DEPTH-1 GENERATE 
	
	m512_wren(i) <= wren WHEN (to_integer(unsigned(wraddress(WRADDR_WIDTH-1 DOWNTO 4))) = i) ELSE '0';  --rts was 5
	
	instantiate_M512s_width : FOR j IN 0 TO NUM_M512S_WIDTH-1 GENERATE  
	
		m512_inst : M512
		PORT MAP (
				data(0)	 => data(0),
				wren	 => m512_wren(i),
				wraddress	 => wrdata(5*j+4 DOWNTO 5*j) & wraddress(3 DOWNTO 0) ,
				rdaddress	 => matchdata(5*j+4 DOWNTO 5*j),
				rden => matchen,
				wrclock => wrclock, 
				rdclock => matchclock, 
				q => single_match_onehot(j)(i*16+15 DOWNTO i*16)
				);

	END GENERATE instantiate_M512s_width;		
END GENERATE instantiate_M512s_depth;

-- For a match, all the M512s must output a 1 on the same bit.  This ANDs them together to find the correct output. For example, if the word is 10 bits long,
-- and we do a match on "00001_00011" then the first 5 bits might match many locations and the second 5 bits might match many locations.  We only want
-- the locations that match both bits(i.e. AND them together)
PROCESS(single_match_onehot)
	variable var_match_onehot : std_logic_vector(NUM_WORDS-1 DOWNTO 0); 
	BEGIN
	var_match_onehot := single_match_onehot(0)(NUM_WORDS-1 DOWNTO 0); -- need to initialize with a 1 the ones that could possibly match up
	FOR k IN 0 TO NUM_M512S_WIDTH-1 LOOP	 
		FOR m IN 0 TO NUM_WORDS-1 LOOP   
			var_match_onehot(m) := var_match_onehot(m) AND single_match_onehot(k)(m);
		END LOOP;
	END LOOP;
	match_onehot <= var_match_onehot;
END PROCESS;
	
-- This process encodes the match output.  So 000...0001000 becomes 11.  (Bit 3 is on, so output a binary 3)
-- Note that this does not do priority encoding.  So if there are multiple matches, this will give erroneous data.	
PROCESS(match_onehot)
	variable code : STD_LOGIC_VECTOR(matchaddr'RANGE);
	BEGIN
    code := (others => '0'); 
    for N in match_onehot'RANGE loop 
      if match_onehot(N) = '1' then 
        code := code OR std_logic_vector(to_unsigned(N, code'LENGTH)); 
      end if; 
    end loop; 
	matchaddr <= code;
END PROCESS;

match <= OR_REDUCE(match_onehot);

end;

