LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE IEEE.STD_LOGIC_MISC.ALL;

-- A single M9K can make a 32 word, 8-bit wide CAM.  This script can be parameterized to chain them together into wider/deeper CAMs.
-- Please keep the depth a multiple of 32, and the width a multiple of 8.  If you don't need a width that is a multiple of 8, just tie off the extra
-- bits so they always match(both when writing and matching)

ENTITY M9K_CAM IS
GENERIC (
	-- These are the parameters modified by the user:
	NUM_WORDS : INTEGER := 128; -- Must be multiple of 32
	WRADDR_WIDTH : INTEGER := 7; -- Enter log2(NUM_WORDS).
	WIDTH : INTEGER := 16 -- Must be multiple of 8

);	
PORT
	(
	-- Write Ports
	wrdata			: IN	std_logic_vector(WIDTH-1 DOWNTO 0);
	wraddress		: IN	std_logic_vector(WRADDR_WIDTH-1 DOWNTO 0);
	wrdelete_n		: IN	std_logic;
	wren			: IN	std_logic;
	wrclock			: IN	std_logic;
	
	-- Match Ports
	matchdata		: IN	std_logic_vector(WIDTH-1 DOWNTO 0);
	matchen			: IN	std_logic;
	matchclock		: IN	std_logic;	
	matchaddr		: OUT 	std_logic_vector(WRADDR_WIDTH-1 DOWNTO 0); -- encoded output.  Remove this if unnecessary, as lot of logic needed to encode.
	match_onehot	: BUFFER	std_logic_vector(NUM_WORDS-1 DOWNTO 0); -- one-hot output
	match			: OUT	std_logic -- Determines if there is a match
	);
end M9K_CAM ;

ARCHITECTURE arch OF M9K_CAM IS

component M9K
	PORT
	(
		data		: IN STD_LOGIC_VECTOR (0 DOWNTO 0);
		rdaddress		: IN STD_LOGIC_VECTOR (7 DOWNTO 0);
		rdclock		: IN STD_LOGIC ;
		rden		: IN STD_LOGIC  := '1';
		wraddress		: IN STD_LOGIC_VECTOR (12 DOWNTO 0);
		wrclock		: IN STD_LOGIC ;
		wren		: IN STD_LOGIC  := '1';
		q		: OUT STD_LOGIC_VECTOR (31 DOWNTO 0)
	);
end component;


CONSTANT NUM_M9KS_DEPTH : INTEGER := NUM_WORDS/32;
CONSTANT NUM_M9KS_WIDTH : INTEGER := WIDTH/8;

SIGNAL data: std_logic_vector(0 DOWNTO 0);
SIGNAL address: std_logic_vector(WIDTH+4 DOWNTO 0);
SIGNAL m9k_wren : STD_LOGIC_VECTOR(NUM_M9KS_DEPTH-1 DOWNTO 0);
TYPE single_match_array IS ARRAY (NUM_M9KS_WIDTH-1 DOWNTO 0) OF STD_LOGIC_VECTOR(NUM_WORDS-1 DOWNTO 0); 
SIGNAL single_match_onehot : single_match_array;

begin

	data(0) <= wrdelete_n;

instantiate_M9Ks_depth : FOR i IN 0 TO NUM_M9KS_DEPTH-1 GENERATE 
	
	m9k_wren(i) <= wren WHEN (to_integer(unsigned(wraddress(WRADDR_WIDTH-1 DOWNTO 5))) = i) ELSE '0';
	
	instantiate_M9Ks_width : FOR j IN 0 TO NUM_M9KS_WIDTH-1 GENERATE  
	
		m9k_inst : M9K
		PORT MAP (
				data(0)	 => data(0),
				wren	 => m9k_wren(i),
--rts				wraddress	 => wrdata(7*j+6 DOWNTO 7*j) & wraddress(4 DOWNTO 0) ,
--rts				rdaddress	 => matchdata(7*j+6 DOWNTO 7*j),
				wraddress	 => wrdata(8*j+7 DOWNTO 8*j) & wraddress(4 DOWNTO 0) ,
				rdaddress	 => matchdata(8*j+7 DOWNTO 8*j),
				rden => matchen,
				wrclock => wrclock, 
				rdclock => matchclock, 
				q => single_match_onehot(j)(i*32+31 DOWNTO i*32)
				);

	END GENERATE instantiate_M9Ks_width;		
END GENERATE instantiate_M9Ks_depth;

-- For a match, all the M9Ks must output a 1 on the same bit.  This ANDs them together to find the correct output. For example, if the word is 16 bits long,
-- and we do a match on "00000001_00000011" then the first 8 bits might match many locations and the second 8 bits might match many locations.  We only want
-- the locations that match both bits(i.e. AND them together)
PROCESS(single_match_onehot)
	variable var_match_onehot : std_logic_vector(NUM_WORDS-1 DOWNTO 0); 
	BEGIN
	var_match_onehot := single_match_onehot(0)(NUM_WORDS-1 DOWNTO 0); -- need to initialize with a 1 the ones that could possibly match up
	FOR k IN 0 TO NUM_M9KS_WIDTH-1 LOOP	 
		FOR m IN 0 TO NUM_WORDS-1 LOOP   
			var_match_onehot(m) := var_match_onehot(m) AND single_match_onehot(k)(m);
		END LOOP;
	END LOOP;
	match_onehot <= var_match_onehot;
END PROCESS;
	
-- This process encodes the match output.  So 000...0001000 becomes 11.  (Bit 3 is on, so output a binary 3)
-- Note that this does not do priority encoding.  So if there are multiple matches, this will give erroneous data.	
PROCESS(match_onehot)
	variable code : STD_LOGIC_VECTOR(matchaddr'RANGE);
	BEGIN
    code := (others => '0'); 
    for N in match_onehot'RANGE loop 
      if match_onehot(N) = '1' then 
        code := code OR std_logic_vector(to_unsigned(N, code'LENGTH)); 
      end if; 
    end loop; 
	matchaddr <= code;
END PROCESS;

match <= OR_REDUCE(match_onehot);

end;

